import inspect
import collections

from .decorator import decorator

__all__ = ["dependency", "get_plugin_instance", "BasePlugin", "Plugin", "CustomPlugin"]

_yaz_plugin_classes = None
_yaz_plugin_instance_cache = {}


@decorator
def dependency(func, **config):
    func.yaz_dependency_config = config
    return func


def get_plugin_list():
    global _yaz_plugin_classes

    if _yaz_plugin_classes is None:
        def get_recursively(cls, plugin_list):
            for plugin in cls.__subclasses__():
                if not plugin.yaz_is_final():
                    plugin_list[plugin.__qualname__].append(plugin)
                get_recursively(plugin, plugin_list)
            return plugin_list

        def get_plugin_type(qualname, plugins):
            classes = sorted(plugins, key=lambda plugin: plugin.yaz_get_ordinal())
            return type(qualname, (Final,) + tuple(classes), {})

        # find all Plugin classes recursively
        plugin_list = get_recursively(BasePlugin, collections.defaultdict(list))

        # combine all classes into their Plugin class (i.e. multiple inherited plugin)
        _yaz_plugin_classes = dict((qualname, get_plugin_type(qualname, plugins))
                                   for qualname, plugins
                                   in plugin_list.items())

    return _yaz_plugin_classes


def get_plugin_class(plugin_class):
    assert issubclass(plugin_class, BasePlugin), type(plugin_class)

    global _yaz_plugin_classes
    classes = get_plugin_list()
    return classes.get(plugin_class.__qualname__)


def get_plugin_instance(plugin_class, *args, **kwargs):
    assert issubclass(plugin_class, BasePlugin), type(plugin_class)

    global _yaz_plugin_instance_cache
    if not plugin_class.__qualname__ in _yaz_plugin_instance_cache:
        plugin_class = get_plugin_class(plugin_class)
        _yaz_plugin_instance_cache[plugin_class.__qualname__] = plugin = plugin_class(*args, **kwargs)

        funcs = [func
                 for _, func
                 in inspect.getmembers(plugin)
                 if inspect.ismethod(func) and hasattr(func, "yaz_dependency_config")]

        for func in funcs:
            signature = inspect.signature(func)
            assert all(parameter.kind is parameter.POSITIONAL_OR_KEYWORD and issubclass(parameter.annotation, BasePlugin) for parameter in signature.parameters.values()), "All parameters for {} must type hint to BasePlugin".format(func)
            func(*[get_plugin_instance(parameter.annotation)
                   for parameter
                   in signature.parameters.values()])

    return _yaz_plugin_instance_cache[plugin_class.__qualname__]


class Final:
    @staticmethod
    def yaz_is_final():
        return True


class BasePlugin:
    @staticmethod
    def yaz_is_final():
        return False

    @staticmethod
    def yaz_get_ordinal():
        # typically used from a yaz_foo_plugin
        return 512


class Plugin(BasePlugin):
    @staticmethod
    def yaz_get_ordinal():
        # typically used from a project configuration
        return 256


class CustomPlugin(Plugin):
    @staticmethod
    def yaz_get_ordinal():
        # typically used from a custom user configuration
        return 128



import argparse
import re


class Parser(argparse.ArgumentParser):
    def __init__(self, *args, **kwargs):
        kwargs["formatter_class"] = argparse.RawDescriptionHelpFormatter
        super().__init__(*args, **kwargs)

    def _format_name(self, name: str, prefix: str = "", postfix: str = ""):
        # todo: cleanup below code.  it does what it needs to do, but is not very pretty
        name = re.sub(r"([^A-Z]+)([A-Z]+)", r"\1-\2", name)
        name = re.sub(r"_", r"-", name)
        name = re.sub(r"-[-]+", r"-", name)
        name = re.sub(r"^[-]+", r"", name)
        name = re.sub(r"[-]+$", r"", name)

        return "".join([prefix, name.lower(), postfix])

    def _add_task(self, parser, task):
        if __debug__:
            from .task import Task
            assert isinstance(parser, Parser)
            assert isinstance(task, Task)

        parameter_map = {}

        for parameter in task.get_parameters():
            args = (self._format_name(parameter.name),)
            kwargs = {
                "help": task.get_configuration("{}__help".format(parameter.name)),
                "choices": task.get_configuration("{}__choices".format(parameter.name)),
            }

            if parameter.default is not parameter.empty:
                args = (self._format_name(parameter.name, "-" if len(parameter.name) == 1 else "--"),)
                if not kwargs["help"]:
                    kwargs["help"] = "defaults to {}={!r}".format(parameter.name, parameter.default)
                kwargs["default"] = parameter.default
                kwargs["dest"] = parameter.name

            type_ = task.get_configuration("{}__type".format(parameter.name), str if parameter.annotation is parameter.empty else parameter.annotation)
            if type_ is bool:
                if parameter.default is parameter.empty:
                    group = parser.add_mutually_exclusive_group(required=True)
                    group.add_argument(
                        self._format_name(parameter.name, "-" if len(parameter.name) == 1 else "--"),
                        dest=parameter.name,
                        action="store_true",
                        help="{}, pass this flag to set to True".format(kwargs["help"])
                    )
                    group.add_argument(
                        self._format_name(parameter.name, "--no-"),
                        dest=parameter.name,
                        action="store_false",
                        help="{}, pass this flag to set to False".format(kwargs["help"])
                    )
                    continue

                else:
                    if parameter.default:
                        args = (self._format_name(parameter.name, "--no-"),)
                        kwargs["action"] = "store_false"
                        kwargs["help"] = "{}, pass this flag to set to False".format(kwargs["help"])
                    else:
                        kwargs["action"] = "store_true"
                        kwargs["help"] = "{}, pass this flag to set to True".format(kwargs["help"])

            elif callable(type_):
                kwargs["type"] = type_

            else:
                raise RuntimeError("Task \"{}\" defines parameter \"{}\" with an unknown type \"{}\"".format(task, parameter, type_))

            parameter_map.update({arg: parameter.name for arg in args})
            parser.add_argument(
                *args,
                **dict((key, value) for key, value in kwargs.items() if not value is None))

        parser.description = task.documentation.full
        parser.set_defaults(yaz_task=task)
        parser.set_defaults(yaz_parameter_map=parameter_map)

    def _get_plugin_documentation(self, tasks):
        first_task = tasks[0][1]
        if not isinstance(first_task, dict):
            docs = [first_task.plugin_documentation.full, ""]

            max_name_length = max(len(name) for name, _ in tasks)
            docs.extend("{:{length}}  {}".format(name, task.documentation.short, length=max_name_length) for name, task in tasks)
            return "\n".join(docs)

    def _add_task_tree_node(self, parser, task):
        while isinstance(task, dict) and len(task) == 1:
            task = next(iter(task.values()))

        if isinstance(task, dict):
            tasks = sorted(task.items())
            subparsers = parser.add_subparsers(description=self._get_plugin_documentation(tasks))
            for name, task in tasks:
                self._add_task_tree_node(
                    subparsers.add_parser(self._format_name(name)),
                    task)

        else:
            self._add_task(parser, task)

    def add_task_tree(self, task_tree):
        assert isinstance(task_tree, dict)
        self._add_task_tree_node(self, task_tree)

    def parse_arguments(self, argv):
        assert isinstance(argv, list)
        assert all(isinstance(arg, str) for arg in argv)
        kwargs = vars(self.parse_args(argv[1:]))
        task = kwargs.pop("yaz_task", None)

        # use the parameter_map to get the original parameter names instead
        # of the names used by the argument parser
        parameter_map = kwargs.pop("yaz_parameter_map", None)
        if parameter_map:
            kwargs = {parameter_map.get(key, key): value for key, value in kwargs.items()}

        return task, kwargs



import importlib
import os.path
import sys


def load(directory_name, module_name):
    """Try to load and return a module

    Will add DIRECTORY_NAME to sys.path and tries to import MODULE_NAME.

    For example:
    load("~/.yaz", "yaz_extension")
    """
    directory_name = os.path.expanduser(directory_name)
    if os.path.isdir(directory_name) and directory_name not in sys.path:
        sys.path.append(directory_name)

    try:
        return importlib.import_module(module_name)
    except ImportError:
        pass



import sys
import unittest

from yaz.parser import Parser
from yaz.task import get_task_tree

from .log import logger


class TestCase(unittest.TestCase):
    def get_caller(self, white_list=None):
        def caller(*arguments):
            try:
                task, kwargs = parser.parse_arguments([sys.argv[0]] + list(arguments))
                logger.debug("Calling %s with %s", task, kwargs)

                self.assertTrue(callable(task), "The parser did not find a task")
                self.assertTrue(isinstance(kwargs, dict), "The parser did not find the parameters")
                return task(**kwargs)

            except SystemExit:
                parser.print_usage()
                raise

        parser = Parser()
        parser.add_task_tree(get_task_tree(white_list))

        return caller



from setuptools import setup

setup(name="yaz",
      packages=["yaz"],
      version="1.0.0",
      description="A scripting tool",
      author="Boudewijn Schoon",
      author_email="yaz@frayja.com",
      url="https://github.com/boudewijn-zicht/yaz",
      license="MIT",
      zip_safe=False,
      test_suite="nose.collector",
      tests_require=["nose"],
      classifiers=[
          "Development Status :: 5 - Production/Stable",
          "Environment :: Console",
          "License :: OSI Approved :: MIT License",
          "Programming Language :: Python :: 3.5",
          "Programming Language :: Python :: 3.6"
      ])
